using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;
using Leap.Unity;
using GPUTools.Common.Scripts.PL.Tools;
using LeapInternal;
using GPUTools.Hair.Scripts.Runtime.Render;
using GPUTools.Hair.Scripts;

namespace Stopper {
    [DefaultExecutionOrder(1000)]
	public class AttachToVertex : MVRScript {

        Atom atom;
        bool connected = false;
        
        DAZSkinV2 skin;
        DAZCharacterRun characterRun;

        DAZClothingItem clothing;
        DAZSkinWrap skinWrap;
        ClothSettings clothSettings;

        DAZHairGroup dazHair;
        HairSimControl hairSimControl;
        DAZBone head;
        int numStrands;

        ComputeBuffer verticesBuffer;
        ComputeBuffer particlesBuffer;
        ComputeBuffer normalsBuffer;

        List<int> triangleIndexes = new List<int>();
        int minVertex, maxVertex;
        int[] triangles;
        Vector2[] uv;

        AsyncGPUReadbackRequest verticesRequest;
        AsyncGPUReadbackRequest normalsRequest;
        AsyncGPUReadbackRequest particlesRequest;

        JSONStorableFloat vertexIndex;

        JSONStorableFloat xOffset;
        JSONStorableFloat yOffset;
        JSONStorableFloat zOffset;

        JSONStorableFloat xRotation;
        JSONStorableFloat yRotation;
        JSONStorableFloat zRotation;

        JSONStorableBool zeroOnReattach;

        Vector3 previousRotation;

        public class ConnectionType {
            public static ConnectionType Skin = new ConnectionType();
            public static ConnectionType Clothing = new ConnectionType();
            public static ConnectionType Hair = new ConnectionType();
        }

        ConnectionType connectionType = ConnectionType.Skin;

        public override void Init() {

            vertexIndex = new JSONStorableFloat("Vertex Index", 0, 0, 0, true, true);
            
			JSONStorableStringChooser atomChooser = new JSONStorableStringChooser(
                "Atom", 
                GetAtomUids(),
                "None", 
                "Atom", 
                (string uid) => {
                    atom = SuperController.singleton.GetAtomByUid(uid);
                    if(atom == null) {
                        Disconnect();
                    }
                });

            RegisterStringChooser(atomChooser);
            var atomChooserPopup = CreateScrollablePopup(atomChooser, false);
            atomChooserPopup.popup.onOpenPopupHandlers += () => {
                atomChooser.choices = GetAtomUids();
            };

            JSONStorableStringChooser connectionTypeChooser = new JSONStorableStringChooser(
                "ItemType", 
                new List<string>(){"Skin", "Clothing", "Hair"},
                "Skin", 
                "Item Type", 
                (string type) => {
                    if(type == "Skin") {
                        connectionType = ConnectionType.Skin;
                    } else if (type == "Clothing") {
                        connectionType = ConnectionType.Clothing;
                    } else if (type == "Hair") {
                        connectionType = ConnectionType.Hair;
                    }
                    Disconnect();
                });
            RegisterStringChooser(connectionTypeChooser);
            var itemTypePopup = CreateScrollablePopup(connectionTypeChooser, false);

            JSONStorableStringChooser itemChooser = new JSONStorableStringChooser(
                    "Item", 
                    GetItemUids(),
                    null, 
                    "Item", 
                    (string uid) => {
                        if(atom == null) {
                            return;
                        }
                        var selector = atom.GetStorableByID("geometry") as DAZCharacterSelector;
                        if(connectionType == ConnectionType.Clothing) {
                            clothing = selector.GetClothingItem(uid);
                            StartCoroutine(ConnectClothing());
                        } else if (connectionType == ConnectionType.Skin && uid == "Skin") {
                            characterRun = selector.GetComponent<DAZCharacterRun>();
                            
                            skin = characterRun.skin;
                            StartCoroutine(ConnectSkin());
                        } else if (connectionType == ConnectionType.Hair) {
                            dazHair = selector.GetHairItem(uid);
                            head = atom.GetStorableByID("head") as DAZBone;
                            StartCoroutine(ConnectHair());
                        }
                    });

            RegisterStringChooser(itemChooser);
            var itemChooserPopup = CreateScrollablePopup(itemChooser, false);
            itemChooserPopup.popup.onOpenPopupHandlers += () => {
                itemChooser.choices = GetItemUids();
            };

            vertexIndex.storeType = JSONStorableParam.StoreType.Full;
            UIDynamicSlider uiSlider = this.CreateSlider(vertexIndex, false);
            uiSlider.rangeAdjustEnabled = false;
            uiSlider.slider.wholeNumbers = true;
            RegisterFloat(vertexIndex);


            vertexIndex.setCallbackFunction = (val) => {
                minVertex = (int)vertexIndex.max;
                maxVertex = 0;
                if(triangles != null) {
                    triangleIndexes.Clear();
                    
                    for(int i = 0; i < triangles.Length; i+=3) {
                        for(int j = 0; j < 3; j++){ 
                            if(triangles[i+j] == (int)val) {
                                if(!triangleIndexes.Contains(i)) {
                                    triangleIndexes.Add(i);
                                }

                                for(int k = 0; k < 3; k++){ 
                                    minVertex = Math.Min(minVertex, triangles[i+k]);
                                    maxVertex = Math.Max(maxVertex, triangles[i+k]);
                                }
                            }
                        }
                    }

                    if(minVertex >= maxVertex) {
                        minVertex = 0;
                        maxVertex = (int)vertexIndex.max;
                    }
                } else {
                    minVertex = 0;
                    maxVertex = (int)vertexIndex.max;
                }
            };

            zeroOnReattach = new JSONStorableBool("zeroLocalOnReattach", false);
            var zeroOnReattachToggle = CreateToggle(zeroOnReattach, true);
            zeroOnReattachToggle.label = "Zero Offset On Reattach";

            var attachToClosest = CreateButton("Re-attach To Closest", true);
            attachToClosest.button.onClick.AddListener(() => AttachToClosest());

            
            xOffset = new JSONStorableFloat("xOffset",0,-1,1,false);
            RegisterFloat(xOffset);
            this.CreateSlider(xOffset, false);
            xOffset.setCallbackFunction = (val) => SyncLocalPosition();

            yOffset = new JSONStorableFloat("yOffset",0,-1,1,false);
            RegisterFloat(yOffset);
            this.CreateSlider(yOffset, false);
            yOffset.setCallbackFunction = (val) => SyncLocalPosition();

            zOffset = new JSONStorableFloat("zOffset",0,-1,1,false);
            RegisterFloat(zOffset);
            this.CreateSlider(zOffset, false);
            zOffset.setCallbackFunction = (val) => SyncLocalPosition();


            CreateTextField(new JSONStorableString("rotationWarning", "Note: modifying rotation sliders directly might yield unpredictable behavior"), false).height = 10;
            xRotation = new JSONStorableFloat("xRotation",0,0,360,true);
            RegisterFloat(xRotation);
            this.CreateSlider(xRotation, false);
            xRotation.setCallbackFunction = (val) => SyncLocalPosition();

            yRotation = new JSONStorableFloat("yRotation",0,-0,360,true);
            RegisterFloat(yRotation);
            this.CreateSlider(yRotation, false);
            yRotation.setCallbackFunction = (val) => SyncLocalPosition();

            zRotation = new JSONStorableFloat("zRotation",0,-0,360,true);
            RegisterFloat(zRotation);
            this.CreateSlider(zRotation, false);
            zRotation.setCallbackFunction = (val) => SyncLocalPosition();
		}

        private List<string> GetAtomUids() {
            List<string> atomUids = new List<string>(){"None"};
            atomUids.AddRange(SuperController.singleton.GetAtoms().Where(a => a.type == "Person").Select(a => a.uid));
            return atomUids;
        }

        private List<string> GetItemUids() {
            List<string> itemUids = new List<string>(){"None"};

            if(atom != null) {
                if(connectionType == ConnectionType.Clothing) {
                    itemUids.AddRange(atom.GetComponentsInChildren<DAZClothingItem>().Select(c => c.uid).ToList());
                } else if(connectionType == ConnectionType.Skin) {
                    itemUids.Add("Skin");
                } else if(connectionType == ConnectionType.Hair) {
                    itemUids.AddRange(atom.GetComponentsInChildren<DAZHairGroup>().Select(c => c.uid).ToList());
                }
            }

            return itemUids;
        }


        private void Disconnect() {
            connected = false;
            verticesBuffer = null;
            normalsBuffer = null;
            triangles = null;
            uv = null;
        }

        private IEnumerator ConnectSkin() {
            Disconnect();
            
            yield return SwapSkin();
            
            connected = true;
            AttachToClosest();

            yield break;
        }

        private IEnumerator SwapSkin() {
            skin = characterRun.skin;
            if(!skin.isActiveAndEnabled) {
                yield return new WaitUntil(() => skin.isActiveAndEnabled);
            }

            yield return null;
            SyncBuffers();
            
            var dazMesh = skin.dazMesh;
            triangles = dazMesh.UVTriangles;
            uv = dazMesh.UV;
            
            vertexIndex.max = verticesBuffer.count-1;
            vertexIndex.setCallbackFunction(vertexIndex.val);
        }

        private IEnumerator ConnectClothing() {
            Disconnect();

            
            if(!clothing.enabled && clothing.ready)
                yield return new WaitUntil(() => clothing.enabled && clothing.ready);

            skinWrap = clothing.GetComponentInChildren<DAZSkinWrap>();
            if(skinWrap == null || !skinWrap.enabled) {
                yield return new WaitUntil(() => {
                    skinWrap = clothing.GetComponentInChildren<DAZSkinWrap>();
                    return skinWrap != null && skinWrap.enabled;
                });
            }

            yield return null;
            
            clothSettings = clothing.GetComponentInChildren<ClothSettings>();
            SyncBuffers();

            var dazMesh = skinWrap.dazMesh;
            triangles = dazMesh.UVTriangles;
            uv = dazMesh.UV;
            
            vertexIndex.max = verticesBuffer.count-1;
            vertexIndex.setCallbackFunction(vertexIndex.val);
          
            connected = true;

            AttachToClosest();

            yield break;
        }

        private IEnumerator ConnectHair() {
            Disconnect();

            if(!dazHair.enabled)
                yield return new WaitUntil(() => dazHair.enabled);

            hairSimControl = dazHair.GetComponentInChildren<HairSimControl>();
            if(hairSimControl == null || !hairSimControl.enabled){
                yield return new WaitUntil(() => {
                    hairSimControl = dazHair.GetComponentInChildren<HairSimControl>();
                    return hairSimControl != null && hairSimControl.enabled;
                });
            }
            yield return null;
            SyncBuffers();
            vertexIndex.max = particlesBuffer.count-2;
            
            vertexIndex.setCallbackFunction(vertexIndex.val);
            connected = true;

            AttachToClosest();

            yield break;
        }

        private void AttachToClosest() {
            if(!connected) {
                return;
            }
            
            float closestDistance = float.MaxValue;
            int closestVertex = 0;
            Vector3 controllerPosition = containingAtom.mainController.transform.position;

            if(connectionType == ConnectionType.Hair) {
                TessRenderParticle[] particles  = new TessRenderParticle[particlesBuffer.count];
                particlesBuffer.GetData(particles);
                for(int i = 0; i < particles.Length; i++) {
                    float sqrDistance = (controllerPosition - particles[i].Position).sqrMagnitude;
                    if(sqrDistance < closestDistance) {
                        closestVertex = i;
                        closestDistance = sqrDistance;
                    }
                }
            } else {
                Vector3[] vertices = new Vector3[verticesBuffer.count];
                verticesBuffer.GetData(vertices);
                for(int i = 0; i < vertices.Length; i++) {
                    float sqrDistance = (controllerPosition - vertices[i]).sqrMagnitude;
                    if(sqrDistance < closestDistance) {
                        closestVertex = i;
                        closestDistance = sqrDistance;
                    }
                }
            }

            vertexIndex.val = closestVertex;

            if(zeroOnReattach.val) {
                xRotation.valNoCallback = 0;
                yRotation.valNoCallback = 0;
                zRotation.valNoCallback = 0;

                xOffset.valNoCallback = 0;
                yOffset.valNoCallback = 0;
                zOffset.valNoCallback = 0;
                SyncLocalPosition();
            }            
        }

        private void SyncBuffers() {
            if(connectionType == ConnectionType.Clothing) {
                if(clothSettings != null && clothSettings.enabled) {
                    verticesBuffer = clothSettings.Runtime.ClothOnlyVertices.ComputeBuffer;
                } else {
                    verticesBuffer = skinWrap.PreCalculatedVerticesBuffer.ComputeBuffer;
                    // Need to call dispatch to copy verts into the pre-compute buffer
                    // Perf hit should be minor
                    skinWrap.Dispatch();
                }
                normalsBuffer = skinWrap.NormalsBuffer.ComputeBuffer;
            } else if (connectionType == ConnectionType.Skin) {
                if(skin != characterRun.skin) {
                    StartCoroutine(SwapSkin());
                }

                if (skin.delayDisplayOneFrame) {
                    verticesBuffer = skin.delayedVertsBuffer;
                } else if (skin.useSmoothing) {
                    verticesBuffer = skin.smoothedVertsBuffer;
                } else {
                    verticesBuffer = skin.rawVertsBuffer;
                }
                normalsBuffer = skin.NormalsBuffer.ComputeBuffer;
          
            } else if (connectionType == ConnectionType.Hair) {
             
                particlesBuffer = hairSimControl.hairSettings.RuntimeData.TessRenderParticles.ComputeBuffer;
                numStrands = hairSimControl.hairSettings.StandsSettings.Provider.GetStandsNum();
            }
        }

        private void SyncLocalPosition() {
            containingAtom.mainController.transform.localPosition = new Vector3(xOffset.val, yOffset.val, zOffset.val);
            containingAtom.mainController.transform.localEulerAngles = new Vector3(xRotation.val, yRotation.val, zRotation.val);
        }

        public void Update() {

        }

        public void FixedUpdate() {

        }

        private bool isConnected() {
            if(!connected)
                return false;

            if(connectionType == ConnectionType.Clothing) {
                if(skinWrap == null) return false;
                if(!skinWrap.enabled) return false;
            } else if(connectionType == ConnectionType.Skin) {
                if(skin == null) return false;
                if(!skin.enabled) return false;
            } else if(connectionType == ConnectionType.Hair) {
                if(hairSimControl == null) return false;
                if(!hairSimControl.enabled) return false;
            }

            return true;        
        }

        public void LateUpdate() {
            
            if(!isConnected()) {
                return;
            }

            var rotation = containingAtom.mainController.transform.localEulerAngles;
            if(rotation != previousRotation) {
                xRotation.valNoCallback = rotation.x;
                yRotation.valNoCallback = rotation.y;
                zRotation.valNoCallback = rotation.z;
            }
            previousRotation = rotation;

            xOffset.valNoCallback = containingAtom.mainController.transform.localPosition.x;
            yOffset.valNoCallback = containingAtom.mainController.transform.localPosition.y;
            zOffset.valNoCallback = containingAtom.mainController.transform.localPosition.z;
            
            SyncBuffers();

            if(connectionType == ConnectionType.Hair) {
                int index = (int)vertexIndex.val;
                if(particlesBuffer != null) {
                    particlesRequest = AsyncGPUReadback.Request(particlesBuffer, TessRenderParticle.Size()*2, index*TessRenderParticle.Size());
                }
                
                if(!particlesRequest.done) {
                    particlesRequest.WaitForCompletion();
                }
                
                if(particlesRequest.hasError) {
                    return;
                }

                var particles = particlesRequest.GetData<TessRenderParticle>();
                var particle = particles[0];
                var nextParticle = particles[1];
                if(particle.Position.ContainsNaN() || nextParticle.Position.ContainsNaN()){
                    return;
                } 

                Vector3 normal;

                if(nextParticle.RootIndex == index+1) { //next particle is in another strand
                    normal = Vector3.down;
                } else {
                    normal = (nextParticle.Position - particle.Position).normalized;
                }
                Vector3 tangent = (head.transform.position - particle.Position).normalized;
                
                containingAtom.reParentObject.position = particle.Position;
                containingAtom.reParentObject.rotation = Quaternion.LookRotation(normal, tangent);

            } else {
                int vertex = (int)vertexIndex.val;
                if(verticesBuffer != null && normalsBuffer != null) {
                    verticesRequest = AsyncGPUReadback.Request(verticesBuffer, (maxVertex-minVertex+1)*12, minVertex*12);
                    normalsRequest = AsyncGPUReadback.Request(normalsBuffer, 12, vertex*12);
                }

                if(!verticesRequest.done) {
                    verticesRequest.WaitForCompletion();
                }
                if(!normalsRequest.done) {
                    normalsRequest.WaitForCompletion();
                }

                if(verticesRequest.hasError || normalsRequest.hasError) {
                    return;
                }

                var vertices = verticesRequest.GetData<Vector3>();
                var normals = normalsRequest.GetData<Vector3>();


                Vector3 position = vertices[vertex-minVertex];
                Vector3 normal = normals[0];

                if(position.ContainsNaN() || normal.ContainsNaN()){
                    return;
                } 
                
                Vector3 tangent = GetTangent(vertex, triangleIndexes, minVertex, triangles, vertices, normals, uv);

                containingAtom.reParentObject.position = position;
                containingAtom.reParentObject.rotation = Quaternion.LookRotation(normal, tangent);
            }
        }

        public static void Log( string str) {
            if(UnityEngine.Time.frameCount % 10 == 0) {
                SuperController.LogMessage(str);
            }
        }

        public JSONStorableFloat AddSlider(string name, float defaultVal, float min, float max, JSONStorableFloat.SetFloatCallback onChange = null, bool rangeAdjust = false, bool rightSide = false){
            JSONStorableFloat storableFloat = new JSONStorableFloat(name, defaultVal, min, max, true, true);
            storableFloat.storeType = JSONStorableParam.StoreType.Full;
            UIDynamicSlider uiSlider = this.CreateSlider(storableFloat, rightSide);
            uiSlider.rangeAdjustEnabled = rangeAdjust;
            uiSlider.valueFormat = "F4";
            RegisterFloat(storableFloat);
            
            if(onChange != null){
                storableFloat.setCallbackFunction = onChange;
            }
            
            return storableFloat;
        }

        // Originally Copied from https://discussions.unity.com/t/calculating-tangents-vector4/179
        public Vector3 GetTangent(int vertex, List<int> triangleIndexes, int vertexOffset, int[] triangles, NativeArray<Vector3> vertices, NativeArray<Vector3> normals, Vector2[] uv) {
            Vector3 tan = new Vector3();

            foreach (int a in triangleIndexes) {
                int i1 = triangles[a + 0];
                int i2 = triangles[a + 1];
                int i3 = triangles[a + 2];

                Vector3 v1 = vertices[i1-vertexOffset];
                Vector3 v2 = vertices[i2-vertexOffset];
                Vector3 v3 = vertices[i3-vertexOffset];

                Vector2 w1 = uv[i1];
                Vector2 w2 = uv[i2];
                Vector2 w3 = uv[i3];

                float x1 = v2.x - v1.x;
                float x2 = v3.x - v1.x;
                float y1 = v2.y - v1.y;
                float y2 = v3.y - v1.y;
                float z1 = v2.z - v1.z;
                float z2 = v3.z - v1.z;

                float s1 = w2.x - w1.x;
                float s2 = w3.x - w1.x;
                float t1 = w2.y - w1.y;
                float t2 = w3.y - w1.y;

                float r = 1.0f / (s1 * t2 - s2 * t1);

                Vector3 sdir = new Vector3((t2 * x1 - t1 * x2) * r, (t2 * y1 - t1 * y2) * r, (t2 * z1 - t1 * z2) * r);

                if(i1 == vertex)
                    tan += sdir;

                if(i2 == vertex)
                    tan += sdir;

                if(i3 == vertex)
                    tan += sdir;;

            }

            
            Vector3 n = normals[0];
            Vector3 t = tan;

            Vector3 tmp = (t - n * Vector3.Dot(n, t)).normalized;
            return new Vector3(tmp.x, tmp.y, tmp.z);

        }
		
		
		private void OnDisable()
		{
			// Clean up maybe? Seems like a hassle.
		}
	}
}